--Functions pertaining to the handling of maps and their data

-- Creates an internal map directory
mapDir = {}
deepfreeze = false

-- Loads the maps into the directory, prints back the name of each, and sets some global values
function loadMaps()
	print("Loading maps...")
	local totesmaps = lf.getDirectoryItems("maps/lua")
	for k, initmap in ipairs(totesmaps) do
		if lf.isFile("maps/lua/"..initmap) then
			print(k..". "..initmap)
			table.insert(mapDir, sti.new("maps/lua/"..initmap))
		end
	end
	print(#mapDir.." map(s) loaded!")
	for i = 1, #mapDir do
		print("o "..mapDir[i].properties["sysname"])
	end
	delay = 120
end

-- Function used to call a new map
function loadMap(mindex, psetx, psety, psetf)
  numKeys = 0
	if not (mindex > #mapDir or mindex <= 0) then
    clearing = false
    canShoot = true
		c_map = mindex
		map = mapDir[mindex]
		print("Loading map "..mindex)
    if map.properties["hasShadows"] then
      print("This map not available on Android!")
    end
		objLayer = map.layers["Objects"] or map.layers["Object Layer 1"] -- the 'or' is to save accidental laziness issues
		mapobjects = convertObjects(objLayer)
		
		if psetx ~= nil and psety ~= nil then
			player.x, player.dx = psetx * 32, psetx * 32
			player.y, player.dy = psety * 32, psety * 32
			camera:setPosition(math.floor(psetx * 32 + 16 - lg.getWidth() / 2), math.floor(psety * 32 + 16 - lg.getHeight() / 2))
		end
		if psetf ~= nil then
			player:moveFace(psetf)
		else
			player:moveFace("down")
		end
	else
		print("Map "..mindex.." not loaded!: Index ["..mindex.."] invalid!")
	end
	
	map:resize(dimsW, dimsH)
  
  stallMap()
	
	function objLayer:draw()
		for i = 1, map.height do
			for k, obj in pairs(mapobjects) do
				if obj.dy <= 0 then
					obj:draw()
				elseif (obj.dy / 32) == i then
					obj:draw()
				end
			end
		end
		lg.setColor(colors.white)
	end
end

-- Function used to convert objects to handleable data
-- returns: newobjlist (a list of 'objects' in readable data)
function convertObjects(oLayer)
	local objs = oLayer.objects
	local newobjlist = {}
	for i = 1, #objs do
		local props = objs[i].properties
		if objs[i].type == "PlayerStart" then
			player = {x = objs[i].x, y = objs[i].y}
			player.name = "Player"
			player.type = "Player"
      player.spriteb = baselist[spritebase]
      if gen == 'm' then
        player.spritec = clotheslist[spriteclothes]
        player.spriteh = hairlist[spritehair]
      elseif gen == 'f' then
        player.spritec = clotheslistf[spriteclothes]
        player.spriteh = hairlistf[spritehair]
      end
			player.dispModX, player.dispModY = 0, 4
			player.xspeed, player.yspeed = 0, 0
			player.speed = 180
			player.face = "down"
			player.dx, player.dy = player.x, player.y
			function player:moveFace(dir)
				if dir == "up" then
					player.spriteb:switch(1, 4, 120)
					player.spritec:switch(1, 4, 120)
					player.spriteh:switch(1, 4, 120)
				elseif dir == "right" then
					player.spriteb:switch(2, 4, 120)
					player.spritec:switch(2, 4, 120)
					player.spriteh:switch(2, 4, 120)
				elseif dir == "down" then
					player.spriteb:switch(3, 4, 120)
					player.spritec:switch(3, 4, 120)
					player.spriteh:switch(3, 4, 120)
				elseif dir == "left" then
					player.spriteb:switch(4, 4, 120)
					player.spritec:switch(4, 4, 120)
					player.spriteh:switch(4, 4, 120)
				end
				player.face = dir
			end
			function player:update(dt)
				if player.x == player.dx and player.y == player.dy and playercontrols then
					if andyup then
						player:moveFace("up")
						if checkCol(player.dx, player.dy, 0, -1) then
							player.yspeed = -player.speed
							player.dy = player.dy - 32
						end
					elseif andyright then
						player:moveFace("right")
						if checkCol(player.dx, player.dy, 1, 0) then
							player.xspeed = player.speed
							player.dx = player.dx + 32
						end
					elseif andydown then
						player:moveFace("down")
						if checkCol(player.dx, player.dy, 0, 1) then
							player.yspeed = player.speed
							player.dy = player.dy + 32
						end
					elseif andyleft then
						player:moveFace("left")
						if checkCol(player.dx, player.dy, -1, 0) then
							player.xspeed = -player.speed
							player.dx = player.dx - 32
						end
					end
				end
				if player.y ~= player.dy or player.x ~= player.dx then
					player.y = player.y + round(player.yspeed * dt) -- just to keep numbers clean...
					player.x = player.x + round(player.xspeed * dt) -- just to keep numbers clean...
					if (player.yspeed < 0 and player.y <= player.dy) or (player.yspeed > 0 and player.y >= player.dy) or (player.xspeed > 0 and player.x >= player.dx) or (player.xspeed < 0 and player.x <= player.dx) then
						player.yspeed = 0
						player.xspeed = 0
						player.y = player.dy
						player.x = player.dx
					end
				end
        for k, spritesb in ipairs(baselist) do
          spritesb:update(dt)
        end
        for k, spritesc in ipairs(clotheslist) do
          spritesc:update(dt)
        end
        for k, spritesh in ipairs(hairlist) do
          spritesh:update(dt)
        end
        for k, spritescf in ipairs(clotheslistf) do
          spritescf:update(dt)
        end
        for k, spriteshf in ipairs(hairlistf) do
          spriteshf:update(dt)
        end
			end
			function player:isMoving()
				if player.y ~= player.dy or player.x ~= player.dx then
					return true
				else
					return false
				end
			end
      function player:hit()
        
      end
			function player:draw()
				if devview == 0 or devview == 1 then
					lg.setColor(colors.white)
					player.spriteb:draw(player.x - player.dispModX, player.y - player.dispModY)
					player.spritec:draw(player.x - player.dispModX, player.y - player.dispModY)
					player.spriteh:draw(player.x - player.dispModX, player.y - player.dispModY)
				end
				if devview == 1 or devview == 2 then
					lg.setColor(colors.playerRed)
					lg.rectangle('line', player.x, player.y, 32, 32)
					sprint(player.face, player.x, player.y - 10)
					sprint(player.dx..", "..player.dy, player.x, player.y - 20)
					sprint(player.name..", "..player.type, player.x, player.y - 30)
				end
			end
			table.insert(newobjlist, player)
		
    elseif objs[i].type == "shot" then
      local shot = {x = objs[i].x, y = objs[i].y}
      shot.dx, shot.dy = objs[i].x, objs[i].y
      shot.type = objs[i].type
      --shot.sprite = nil
      shot.xspeed, shot.yspeed = 0, 0
      shot.speed = 360
      shot.face = props["face"] or "down"
      function shot:update(dt)
        if shot.y ~= shot.dy or shot.x ~= shot.dx then
					shot.y = shot.y + round(shot.yspeed * dt) -- just to keep numbers clean...
					shot.x = shot.x + round(shot.xspeed * dt) -- just to keep numbers clean...
					if (shot.yspeed < 0 and shot.y <= shot.dy) or (shot.yspeed > 0 and shot.y >= shot.dy) or (shot.xspeed > 0 and shot.x >= shot.dx) or (shot.xspeed < 0 and shot.x <= shot.dx) then
						shot.yspeed = 0
						shot.xspeed = 0
						shot.y = shot.dy
						shot.x = shot.dx
					end
        else
          if shot.face == "down" then
            shot.yspeed = 1200
            shot.dy = shot.dy + 32
          end
				end
				--shot.sprite:update(dt)
      end
      function shot:draw()
        lg.rectangle('line', shot.x, shot.y, 32, 32)
      end
      table.insert(newobjlist, shot)
		
		elseif objs[i].type == "NPC" then
			local npc = {x = objs[i].x, y = objs[i].y}
			npc.dx, npc.dy = objs[i].x, objs[i].y
			npc.name = objs[i].name
			npc.type = objs[i].type
      if props["pattern"] then
        npc.pattern = Tserial.unpack(props["pattern"])
      else
        npc.pattern = { }
      end
			npc.behavior = props["behavior"] or nil
			npc.bspec = props["behaviorspec"] or nil
			npc.frame = 0
			npc.delay = props["delay"] or math.random(3, 10)
			npc.delta = 0
			npc.maxdelta = tonumber(props["maxdelta"]) or math.random(15, 25)
			npc.stalled = props["stalled"] or false
      npc.waiting = false
			npc.static = props["static"] or false
			npc.off = props["off"] or false
      npc.kills = (not props["kills"])
      npc.killable = props["killable"]
      npc.maxhp = tonumber(props["hp"]) or 2
      npc.hp = npc.maxhp
      npc.stallable = props["stallable"]
			npc.sprite = SpriteAnimation:new(chrPath..(props["spriteSheet"] or "null")..".png", (props["gW"] or 32), (props["gH"] or 36), 4, 4)
      if props["color"] then
        npc.color = Tserial.unpack(props["color"])
      else
        npc.color = colors.white
      end
			if props["gW"] then
				npc.dispModX = (props["gW"] - 32) / 2
			else
				npc.dispModX = 0
			end
			if props["gH"] then
				npc.dispModY = (props["gH"] - 32)
			else
				npc.dispModY = 4
			end
			npc.xspeed, npc.yspeed = 0, 0
			npc.speed = tonumber(props["speed"]) or 180
			npc.face = props["face"] or "down"
			npc.oldFace = props["face"] or "down"
			function npc:update(dt)
				-- Updates NPC movement behaviours
				if not npc.off then
					if not npc.stalled then
						npc.delta = npc.delta + npc.delay * dt
						
						if npc.delta >= npc.maxdelta then
							if npc.y == npc.dy and npc.x == npc.dx then
								if npc.frame >= #npc.pattern then
									npc.frame = 1
								else
									npc.frame = npc.frame + 1
								end
								if npc.pattern[npc.frame] == 1 then
									npc:move("up")
								elseif npc.pattern[npc.frame] == 2 then
									npc:move("right")
								elseif npc.pattern[npc.frame] == 3 then
									npc:move("down")
								elseif npc.pattern[npc.frame] == 4 then
									npc:move("left")
								end
							end
							npc.delta = 0
						end
					end
					
					if npc.y ~= npc.dy or npc.x ~= npc.dx then
						npc.y = npc.y + round(npc.yspeed * dt)
						npc.x = npc.x + round(npc.xspeed * dt)
						if (npc.yspeed < 0 and npc.y <= npc.dy) or (npc.yspeed > 0 and npc.y >= npc.dy) or (npc.xspeed > 0 and npc.x >= npc.dx) or (npc.xspeed < 0 and npc.x <= npc.dx) then
							npc.yspeed = 0
							npc.xspeed = 0
							npc.y = npc.dy
							npc.x = npc.dx
						end
					end
					npc.sprite:update(dt)
				else
					npc.x, npc.y = -32, -32
					npc.dx, npc.dy = -32, -32
				end
			end
			function npc:move(dir, mface)
				if dir == nil then
					print("NPC: "..npc.name.." has no move() instruction!")
				elseif dir == "up" and checkCol(npc.dx, npc.dy, 0, -1, true, npc.kills) then
					npc.yspeed = -npc.speed
					npc.dy = npc.dy - 32
					if not mface then
						npc:moveface(dir)
					end
				elseif dir == "right" and checkCol(npc.dx, npc.dy, 1, 0, true, npc.kills) then
					npc.xspeed = npc.speed
					npc.dx = npc.dx + 32
					if not mface then
						npc:moveface(dir)
					end
				elseif dir == "down" and checkCol(npc.dx, npc.dy, 0, 1, true, npc.kills) then
					npc.yspeed = npc.speed
					npc.dy = npc.dy + 32
					if not mface then
						npc:moveface(dir)
					end
				elseif dir == "left" and checkCol(npc.dx, npc.dy, -1, 0, true, npc.kills) then
					npc.xspeed = -npc.speed
					npc.dx = npc.dx - 32
					if not mface then
						npc:moveface(dir)
					end
				end
			end
			function npc:moveface(dir)
				if dir == "up" then
					npc.sprite:switch(1, 4, 120)
				elseif dir == "right" then
					npc.sprite:switch(2, 4, 120)
				elseif dir == "down" then
					npc.sprite:switch(3, 4, 120)
				elseif dir == "left" then
					npc.sprite:switch(4, 4, 120)
				elseif dir == "player" and (math.abs(npc.dy - player.dy) > math.abs(npc.dx - player.dx)) then
					if player.dy > npc.dy then
						npc:moveface("down")
					elseif player.dy < npc.dy then
						npc:moveface("up")
					end
				elseif dir == "player" then
					if player.dx > npc.dx then
						npc:moveface("right")
					elseif player.dx < npc.dx then
						npc:moveface("left")
					end
				end
				if dir ~= "player" then
					npc.face = dir
				end
			end
			function npc:detpattern()
				if npc.behavior then
          if npc.behavior == "pattern" then
            -- Do nothing...
					elseif npc.behavior == "random" then
						for i = 1, 42 do
							npc.pattern[i] = math.random(1, 4)
						end
					elseif npc.behavior == "circle" then
						if npc.bspec then
							if npc.bspec == 0 or npc.bspec == 1 then
								npc.pattern = {1, 2, 3, 4}
							elseif npc.bspec == 2 then
								npc.pattern = {1, 1, 2, 2, 3, 3, 4, 4}
							elseif npc.bspec == 3 then
								npc.pattern = {1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4}
							elseif npc.bspec >= 4 then
								npc.pattern = {1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4}
							end
						else
							npc.behavior = nil
						end
					end
				end
			end
			function npc:isMoving()
				if npc.y ~= npc.dy or npc.x ~= npc.dx then
					return true
				else
					return false
				end
			end
			function npc:draw()
				if not npc.off then
					if devview == 0 or devview == 1 then
						lg.setColor(npc.color)
						npc.sprite:draw(npc.x - npc.dispModX, npc.y - npc.dispModY)
					end
					if devview == 1 or devview == 2 then
						lg.setColor(colors.npcBlue)
						lg.rectangle('line', npc.x, npc.y, 32, 32)
						sprint(npc.face, npc.x, npc.y - 10)
						sprint(npc.dx..", "..npc.dy, npc.x, npc.y - 20)
						sprint(npc.name..", "..npc.type, npc.x, npc.y - 30)
					end
				end
			end
      function npc:movewithplayer()
        if not npc:isMoving() and npc.waiting then
          if player.dy > npc.dy and checkCol(npc.dx, npc.dy, 0, -1, false, false, true) then
            print("I would move up")
            npc.dy = npc.dy - 32
            npc.yspeed = -player.speed
          elseif player.dy < npc.dy and checkCol(npc.dx, npc.dy, 0, 1, false, false, true) then
            print("I would move down")
            npc.dy = npc.dy + 32
            npc.yspeed = player.speed
          elseif player.dx > npc.dx and checkCol(npc.dx, npc.dy, -1, 0, false, false, true) then
            print("I would move left")
            npc.dx = npc.dx - 32
            npc.xspeed = -player.speed
          elseif player.dx < npc.dx and checkCol(npc.dx, npc.dy, 1, 0, false, false, true) then
            print("I would move right")
            npc.dx = npc.dx + 32
            npc.xspeed = player.speed
          end
        end
      end
      function npc:hit()
        npc.color = colors.red
        npc.hp = npc.hp - 1
        if not npc.waiting and npc.stallable then
          print(npc.name.." hit!")
          if npc.stalled then
            print("will stall")
            npc.stallagain = true
          else
            npc.stallagain = false
          end
          npc.stalled = true
          npc.waiting = true
          Timer.add(7, function() if npc.stallagain == true then npc.stalled = true else npc.stalled = false end npc.waiting = false npc.hp = npc.maxhp npc.color = colors.white end)
        elseif npc.killable and npc.hp <= 0 then
          la.play(deathsound)
          npc.off = true
        end
      end
			npc:moveface((props["face"] or "down"))
			npc:detpattern()
			npc.sprite:load(delay)
			table.insert(newobjlist, npc)
			
		elseif objs[i].type == "Counter" then
			local counter = {x = objs[i].x, y = objs[i].y}
			counter.dx, counter.dy = counter.x, counter.y
			counter.type = objs[i].type
			function counter:update(dt)
				-- Do nothing
			end
			function counter:draw()
				if devview == 1 or devview == 2 then
					lg.setColor(colors.counterBrown)
					lg.rectangle('line', counter.x, counter.y, 32, 32)
					sprint(counter.type, counter.x, counter.y - 20)
					sprint(counter.dx..", "..counter.dy, counter.x, counter.y - 10)
				end
			end
			table.insert(newobjlist, counter)
		
		elseif objs[i].type == "Transfer" then
			local transObj = {x = objs[i].x, y = objs[i].y}
			transObj.dx, transObj.dy = objs[i].x, objs[i].y
			transObj.name = objs[i].name
			transObj.type = objs[i].type
			transObj.sendto = props["sendto"]
			transObj.newx = props["newx"] or nil
			transObj.newy = props["newy"] or nil
			transObj.newf = props["newf"] or nil
			transObj.off = props["off"] or false
			function transObj:update(dt)
				-- Do nothing
				if transObj.off then
					transObj.x, transObj.y = -32, -32
					transObj.dx, transObj.dy = -32, -32
				end
			end
			function transObj:draw()
				if devview == 1 or devview == 2 then
					lg.setColor(colors.transferVilt)
					lg.rectangle('line', transObj.x, transObj.y, 32, 32)
					sprint(transObj.name..", "..transObj.type, transObj.x, transObj.y - 20)
					sprint(transObj.dx..", "..transObj.dy, transObj.x, transObj.y - 10)
				end
			end
			table.insert(newobjlist, transObj)
		
		elseif objs[i].type == "Solid" then
			local solidObj = {x = objs[i].x, y = objs[i].y}
			solidObj.dx, solidObj.dy = objs[i].x, objs[i].y
			solidObj.name = objs[i].name
			solidObj.type = objs[i].type
      solidObj.subtype = props["subtype"]
      solidObj.giveShots = props["givesShots"] or 0
      solidObj.chestopen = false
			solidObj.off = props["off"] or false
      solidObj.kills = props["kills"] or false
      solidObj.moveable = props["moveable"]
      solidObj.drawn = props["drawn"]
      solidObj.speed = player.speed
      solidObj.xspeed, solidObj.yspeed = 0, 0
      if props["sprite"] ~= nil and lf.isFile(sldPath..props["sprite"]..".png") then
        solidObj.sprite = lg.newImage(sldPath..props["sprite"]..".png")
      else
        solidObj.sprite = nil
      end
      if solidObj.subtype == "key" then
        local sprnum = round(math.random(1, 3))
        solidObj.sprite = lg.newImage(sldPath.."book"..sprnum..".png")
      end
			function solidObj:update(dt)
				if not solidObj.off then
					if solidObj.y ~= solidObj.dy or solidObj.x ~= solidObj.dx then
            if solidObj.y < solidObj.dy then
              solidObj.yspeed = solidObj.speed
            elseif solidObj.y > solidObj.dy then
              solidObj.yspeed = -solidObj.speed
            end
            if solidObj.x < solidObj.dx then
              solidObj.xspeed = solidObj.speed
            elseif solidObj.x > solidObj.dx then
              solidObj.xspeed = -solidObj.speed
            end
						solidObj.y = solidObj.y + round(solidObj.yspeed * dt)
						solidObj.x = solidObj.x + round(solidObj.xspeed * dt)
						if (solidObj.yspeed < 0 and solidObj.y <= solidObj.dy) or (solidObj.yspeed > 0 and solidObj.y >= solidObj.dy) or (solidObj.xspeed > 0 and solidObj.x >= solidObj.dx) or (solidObj.xspeed < 0 and solidObj.x <= solidObj.dx) then
							solidObj.yspeed = 0
							solidObj.xspeed = 0
							solidObj.y = solidObj.dy
							solidObj.x = solidObj.dx
						end
					end
				else
					solidObj.x, solidObj.y = -32, -32
					solidObj.dx, solidObj.dy = -32, -32
				end
			end
      function solidObj:isMoving()
				if solidObj.y ~= solidObj.dy or solidObj.x ~= solidObj.dx then
					return true
				else
					return false
				end
			end
			function solidObj:draw()
				if devview == 1 or devview == 2 then
					lg.setColor(colors.solidGray)
					lg.rectangle('line', solidObj.x, solidObj.y, 32, 32)
					sprint(solidObj.name..", "..solidObj.type, solidObj.x, solidObj.y - 20)
					sprint(solidObj.dx..", "..solidObj.dy, solidObj.x, solidObj.y - 10)
        elseif solidObj.drawn then
          if solidObj.sprite ~= nil then
            lg.setColor(colors.white)
            if solidObj.subtype == "door" then
              lg.setColor(255, 255, 255, 175)
            end
            lg.draw(solidObj.sprite, solidObj.x, solidObj.y)
          else
            lg.setColor(colors.solidGray)
            lg.rectangle('line', solidObj.x, solidObj.y, 32, 32)
          end
				end
			end
      function solidObj:open()
        if solidObj.subtype == "chest" then
          print("I'm open!")
          solidObj.sprite = lg.newImage(sldPath.."chest_o.png")
          solidObj.chestopen = true
          la.play(boxsound)
        end
      end
      function solidObj:movewithplayer()
        if solidObj.subtype == "chest" and solidObj.chestopen == true then
          solidObj.sprite = lg.newImage(sldPath.."chest_e.png")
          clearMap()
          la.play(doorsound)
        end
        if solidObj.subtype == "portal" then
          stagecall = props["portalto"]
          la.play(movesound)
        end
        if solidObj.subtype == "key" then
          numKeys = numKeys - 1
          numShots = numShots + solidObj.giveShots
          solidObj.off = true
          solidObj.drawn = false
          la.play(keysound)
        end
        if not solidObj:isMoving() and solidObj.subtype ~= "chest" and solidObj.subtype ~= "portal" and solidObj.subtype ~= "key" then
          if player.dy > solidObj.dy and checkCol(solidObj.dx, solidObj.dy, 0, -1, false, false, true) then
            print("I would move up")
            solidObj.dy = solidObj.dy - 32
          elseif player.dy < solidObj.dy and checkCol(solidObj.dx, solidObj.dy, 0, 1, false, false, true) then
            print("I would move down")
            solidObj.dy = solidObj.dy + 32
          elseif player.dx > solidObj.dx and checkCol(solidObj.dx, solidObj.dy, -1, 0, false, false, true) then
            print("I would move left")
            solidObj.dx = solidObj.dx - 32
          elseif player.dx < solidObj.dx and checkCol(solidObj.dx, solidObj.dy, 1, 0, false, false, true) then
            print("I would move right")
            solidObj.dx = solidObj.dx + 32
          end
        end
      end
      function solidObj:hit()
        
      end
      if solidObj.subtype == "key" then
        numKeys = numKeys + 1
      end
			table.insert(newobjlist, solidObj)
		
		elseif objs[i].type == "Floor (No NPC)" then
			local floorObj = {x = objs[i].x, y = objs[i].y}
			floorObj.dx, floorObj.dy = objs[i].x, objs[i].y
			floorObj.name = objs[i].name
			floorObj.type = objs[i].type
			function floorObj:update(dt)
				-- Do nothing
			end
			function floorObj:draw()
				if devview == 1 or devview == 2 then
					lg.setColor(colors.floorNoNPC)
					lg.rectangle('line', floorObj.x, floorObj.y, 32, 32)
					sprint(floorObj.name..", "..floorObj.type, floorObj.x, floorObj.y - 20)
					sprint(floorObj.dx..", "..floorObj.dy, floorObj.x, floorObj.y - 10)
				end
			end
			table.insert(newobjlist, floorObj)
		end
	end
	return newobjlist
end

-- Finder Functions

-- Searches for an object by name
function findObjNamed(joe, ret_obj)
	for i = 1, #mapobjects do
		if mapobjects[i].name == joe then
			if ret_obj then
				return mapobjects[i]
			else
				return i
			end
		end
	end
end

-- Searches for a map by name
function findMapNamed(joe)
	for i = 1, #mapDir do
		if mapDir[i].properties["sysname"] == joe then
			return i
		end
	end
end

-- Searches for a state by map name
function findStateNamed(joe)
	
end

-- Checks all cards for collision with mouse
function checkMousePos()
	cardsLayer = map.layers["Cards"]
	local mx, my = lm.getPosition()
	local widthoff = (-176 + lw.getWidth() / 2)
	local heightoff = (-176 + lw.getHeight() / 2)
	for i = 1, tonumber(cardsLayer.properties["tot"]) do
		cardData = cardsLayer.objects[i]
		if mx >= cardData.x + widthoff and mx <= (cardData.x + cardData.width + widthoff) and my >= cardData.y + heightoff and my <= (cardData.y + cardData.height + heightoff) then
			print(i)
		end
	end
end

function getMousePos(mpx, mpy, noffset)
	cardsLayer = map.layers["Cards"]
  local widthoff, heightoff = 0, 0
  if noffset then
    widthoff = (-272 + lw.getWidth() / 2)
    heightoff = (-208 + lw.getHeight() / 2)
  else
    widthoff = (-176 + lw.getWidth() / 2)
    heightoff = (-176 + lw.getHeight() / 2)
  end
	for i = 1, 14 do
		cardData = cardsLayer.objects[i]
		if cardData ~= nil and mpx >= cardData.x + widthoff and mpx <= (cardData.x + cardData.width + widthoff) and mpy >= cardData.y + heightoff and mpy <= (cardData.y + cardData.height + heightoff) then
			return i
		end
	end
end

-- Movement Functions
function stallMap()
  for i = 1, #mapobjects do
    if mapobjects[i].type == "NPC" then
      mapobjects[i].stalled = true
    end
  end
end

function unstallMap()
  for i = 1, #mapobjects do
    if mapobjects[i].type == "NPC" then
      if mapobjects[i].waiting then
        mapobjects[i].stallagain = false
      else
        mapobjects[i].stalled = false
      end
    end
  end
end

function freezeMap(waitMotionless)
	playercontrols = false
	for i = 1, #mapobjects do
		if mapobjects[i].type == "NPC" then
			mapobjects[i].stalled = true
		end
	end
	deepfreeze = true
	if waitMotionless then
		waitTilMotionless()
	end
end

function unfreezeMap()
	playercontrols = true
	for i = 1, #mapobjects do
		if mapobjects[i].type == "NPC" then
			mapobjects[i].stalled = false
		end
	end
	deepfreeze = false
end

function clearMap()
  local killval = #mapobjects
  while killval > 0 do
    if mapobjects[killval].type == "NPC" or mapobjects[killval].subtype == "door" or mapobjects[killval].type == "shot" then
      print("removing "..mapobjects[killval].type)
      table.remove(mapobjects, killval)
    end
    killval = killval - 1
  end
end

function waitTilMotionless()
	for i = 1, #mapobjects do
		if (mapobjects[i].type == "NPC" or mapobjects[i].type == "Player") and mapobjects[i]:isMoving() then
			i = 1
		end
	end
	print("Motionless!")
end

function killPlayer()
  print("YOU DIED~!")
  la.play(deathsound)
  State.switch(c_state)
  Timer.add(.5, function() loadMap(c_map) end)
  stallMap()
end

-- Checks collision of the next tile
-- returns: (if object can move), (what's blocking it)
function checkCol(cdx, cdy, x, y, npcChecking, killer, withplayer)
    local nextX, nextY = (cdx / 32) + x, (cdy / 32) + y
    local layer = map:getCollisionMap("Collision")
    local tile = layer.data[nextY + 1][nextX + 1]
    if killer == nil then
      killer = false
    end
    if withplayer == nil then
      withplayer = false
    end
    if tile == 1 then
        return false, nil
    end
    for i = 1, #mapobjects do
		if (mapobjects[i].dx / 32) == nextX and (mapobjects[i].dy / 32) == nextY  then
			if mapobjects[i].type == "NPC" or mapobjects[i].type == "Solid" then
        if (mapobjects[i].moveable or mapobjects[i].waiting) and not npcChecking and not withplayer then
          mapobjects[i].movewithplayer()
        end
				return false, mapobjects[i]
			elseif (mapobjects[i].type == "Player" or mapobjects[i].type == "Floor (No NPC)" or mapobjects[i].type == "Transfer") and npcChecking then
        if mapobjects[i].type == "Player" and killer then
          killPlayer()
        end
				return false, mapobjects[i]
			end
		end
    end
    return true, nil
end

-- Checks near-by object for actions
function checkAct(dxmod, dymod)
	if dxmod == nil then
		dxmod = 0
	end
	if dymod == nil then
		dymod = 0
	end
	for i = 1, #mapobjects do
		if i > #mapobjects then
			return
		end
		if mapobjects[i].type == "NPC" or mapobjects[i].type == "Solid" then
			if player.face == "up" and (mapobjects[i].dx - dxmod == player.dx and mapobjects[i].dy - dymod == player.dy - 32) then
				msgDraw = mapobjects[i].name
			elseif player.face == "down" and (mapobjects[i].dx - dxmod == player.dx and mapobjects[i].dy - dymod == player.dy + 32) then
				msgDraw = mapobjects[i].name
			elseif player.face == "right" and (mapobjects[i].dx - dxmod == player.dx + 32 and mapobjects[i].dy - dymod == player.dy) then
				msgDraw = mapobjects[i].name
			elseif player.face == "left" and (mapobjects[i].dx - dxmod == player.dx - 32 and mapobjects[i].dy - dymod == player.dy) then
				msgDraw = mapobjects[i].name
			end
		elseif mapobjects[i].type == "Counter" then
			if player.face == "up" and (mapobjects[i].dx - dxmod == player.dx and mapobjects[i].dy - dymod == player.dy - 32) then
				checkAct(0, -32)
				print("Counter obj: Checking...")
				print("...up!")
			elseif player.face == "down" and (mapobjects[i].dx - dxmod == player.dx and mapobjects[i].dy - dymod == player.dy + 32) then
				checkAct(0, 32)
				print("Counter obj: Checking...")
				print("...down!")
			elseif player.face == "right" and (mapobjects[i].dx - dxmod == player.dx + 32 and mapobjects[i].dy - dymod == player.dy) then
				checkAct(32, 0)
				print("Counter obj: Checking...")
				print("...right!")
			elseif player.face == "left" and (mapobjects[i].dx - dxmod == player.dx - 32 and mapobjects[i].dy - dymod == player.dy) then
				checkAct(-32, 0)
				print("Counter obj: Checking...")
				print("...left!")
			end
		elseif mapobjects[i].type == "Transfer" then
			if mapobjects[i].dx == player.dx and mapobjects[i].dy == player.dy then
				print("Transfer obj: Sending to "..mapobjects[i].sendto)
				msgDraw = mapobjects[i].sendto
				Timer.add(.5, function() loadMap(findMapNamed(mapobjects[i].sendto), mapobjects[i].newx, mapobjects[i].newy, mapobjects[i].newf) end)
			end
		end
	end
end

-- System Functions

-- Runs the update function of each object
function updateObjs(dt)
	if not ingamemenu then
		for i = 1, #mapobjects do
      if mapobjects[i] then
        mapobjects[i]:update(dt)
      end
		end
	end
end

function callTransfer(sendaway, newx, newy, newf)
	Timer.add(.5, function() loadMap(findMapNamed(sendaway), newx, newy, newf) end)
end

function objLocsToTable(ret)
	if ret == nil then
		ret = false
	end
	freezeMap(true)
	local tblx, tbly, tbloff, tblf = { }, { }, { }, { }
	for i = 1, #mapobjects do
		table.insert(tblx, mapobjects[i].dx)
		table.insert(tbly, mapobjects[i].dy)
		table.insert(tbloff, mapobjects[i].off)
		table.insert(tblf, mapobjects[i].face)
	end
	local retx = Tserial.pack(tblx)
	local rety = Tserial.pack(tbly)
	local retoff = Tserial.pack(tbloff)
	local retf = Tserial.pack(tblf)
	local retstr = ("-- X's\r\n"..retx.."\r\n-- Y's\r\n"..rety.."\r\n-- Off?\r\n"..retoff.."\r\n-- Face\r\n"..retf)
	if ret then
		print("Returned locData!")
		return retstr
	else
		print("Written locData!")
		lf.write(locData..map.properties["sysname"]..".lcd", retstr)
	end
	unfreezeMap()
end

function objTableToLocs(playerstatic)
	if lf.exists(locData..map.properties["sysname"]..".lcd") then
		if playerstatic == nil then
			playerstatic = false
		end
		freezeMap(true)
		local tbls = { }
		for xys in lf.lines(locData..map.properties["sysname"]..".lcd") do
			table.insert(tbls, xys)
		end
		local tblx = Tserial.unpack(tbls[2])
		local tbly = Tserial.unpack(tbls[4])
		local tbloff = Tserial.unpack(tbls[6])
		for i = 1, #mapobjects do
			if mapobjects[i].type == "Player" and playerstatic then
				-- Do nothing
				print("Not loading player locData!")
			elseif not mapobjects[i].static then
				mapobjects[i].dx = tblx[i]
				mapobjects[i].x = tblx[i]
				mapobjects[i].dy = tbly[i]
				mapobjects[i].y = tbly[i]
				mapobjects[i].off = tbloff[i]
			end
		end
		unfreezeMap()
		print("Loaded locData!")
	else
		print("No locs tables!")
	end
	staticTransfer = true
end